#include "PX_Firework.h"
#include "math.h"

#define PX_FIREWORKLAUNCHER_COUNT 128
#define PX_FIREWORKLAUNCHER_MAX_INTERPOLATE_COUNT 128
#define PX_FIREWORKLAUNCHER_MAX_PARTICAL 64

#define PX_FIREWORKLAUNCHER_DURATION 200
#define PX_FIREWORKLAUNCHER_GEN_DURATION 50
#define PX_FIREWORKLAUNCHER_NORMAL_TIME 4000
#define PX_FIREWORKLAUNCHER_FLOWER_TIME 2000
#define PX_FIREWORKBLOOM_TIME 5000
px_bool bRun = 0;

typedef struct
{
	px_point array[32];
	px_dword delay;
	px_float size;
	px_float vx, vy;
	px_color clr;
}PX_FireworkBloom_Partical;

typedef struct
{
	PX_FireworkBloom_Partical particals[96];
	px_bool activate;
	px_dword elapsed;
}PX_FireworkBloom;

typedef enum 
{
	PX_FIREWORK_LEAP_TYPE_NORMAL,
	PX_FIREWORK_LEAP_TYPE_FLOWER,
}PX_FIREWORK_LEAP_TYPE;

typedef enum
{
	PX_FIREWORK_LEAP_TYPE_RAISING,
	PX_FIREWORK_LEAP_TYPE_BLOOMING,
}PX_FIREWORK_LEAP_FLOWER_STATE;

typedef struct
{
	px_float x, y, size;
	px_point4D color;
}PX_FireworkLeapInterpolate;

typedef struct
{
	px_float x, y, size;
	px_float vx, vy;
	px_point4D color;
	px_bool activate;
}PX_FireworkLeapPartical;


typedef struct  
{
	PX_FIREWORK_LEAP_FLOWER_STATE flowerState;
	PX_FIREWORK_LEAP_TYPE type;
	px_float x,y,vx,vy;
	px_point4D src_color,dst_color;
	px_float size;
	px_int reg_interpolate;
	px_int reg_ipartical;
	px_dword reg_elapsed;
	
	PX_FireworkLeapInterpolate interpolate[PX_FIREWORKLAUNCHER_MAX_INTERPOLATE_COUNT];
	PX_FireworkLeapPartical leapparticals[PX_FIREWORKLAUNCHER_MAX_PARTICAL];
	PX_FireworkBloom Bloom;
	px_int alive;
	px_dword elapsed;
}PX_FireworkLeapLauncher;

px_int counter;
px_dword launcher_elapsed;
PX_FireworkLeapLauncher LeapLauncher[PX_FIREWORKLAUNCHER_COUNT] = {0};


px_void PX_FireworkBloomParticalUpdate(PX_FireworkBloom_Partical* pdesc, px_dword elapsed)
{
	px_int i;
	px_dword update_atom;
	if (elapsed > 1000)
	{
		elapsed = 1000;
	}
	while (elapsed)
	{
		if (elapsed > 10)
		{
			update_atom = 10;
			elapsed -= 10;
		}
		else
		{
			update_atom = elapsed;
			elapsed = 0;
		}

		pdesc->array[0].x += pdesc->vx * update_atom / 1000;
		pdesc->array[0].y += pdesc->vy * update_atom / 1000;

		pdesc->vx -= pdesc->vx * 0.9f * update_atom / 1000;
		pdesc->vy += -pdesc->vy * 0.9f * update_atom / 1000;

		pdesc->delay += update_atom;


		while (pdesc->delay > 20)
		{
			pdesc->delay -= 20;
			for (i = PX_COUNTOF(pdesc->array) - 1; i > 0; i--)
			{
				pdesc->array[i] = pdesc->array[i - 1];
			}
		}
	}



}

px_void PX_FireworkBloomRender(px_surface* psurface, PX_FireworkBloom* pBloom, px_dword elapsed)
{
	px_int i,k;
	px_float scale;
	px_color color;

	pBloom->elapsed += elapsed;
	if (pBloom->elapsed> PX_FIREWORKBLOOM_TIME)
	{
		return;
	}
	for (i = 0; i < PX_COUNTOF(pBloom->particals); i++)
	{
		PX_FireworkBloom_Partical* pdesc = pBloom->particals+i;
		PX_FireworkBloomParticalUpdate(pdesc, elapsed);
		for (k = PX_COUNTOF(pdesc->array) - 1; k >= 0; k--)
		{
			color = pdesc->clr;
			color._argb.a = (px_uchar)((PX_FIREWORKBLOOM_TIME * 1.0f - pBloom->elapsed) / PX_FIREWORKBLOOM_TIME * 255);
			scale = (PX_COUNTOF(pdesc->array) - k) * (0.5f + elapsed / 2000.f) / PX_COUNTOF(pdesc->array);

			color._argb.a = (px_uchar)(color._argb.a * scale);
			PX_GeoDrawSolidCircle(psurface, (int)pdesc->array[k].x, (int)pdesc->array[k].y, (int)(pdesc->size * scale), color);
		}
	}
}

px_void PX_FireworkBloomCreate(PX_FireworkBloom* pdesc,px_float x, px_float y)
{
	px_int i;
	px_int j;
	px_color clr;
	px_int randv=PX_randRange(0,20);
	pdesc->elapsed = 0;

	for (i = 0; i < PX_COUNTOF(pdesc->particals); i++)
	{
		for (j = 0; j < PX_COUNTOF(pdesc->particals->array); j++)
		{
			pdesc->particals[i].array[j].x = x;
			pdesc->particals[i].array[j].y = y;
		}
	}
	clr = PX_COLOR(255, 128 + (PX_rand() % 128), 128 + (PX_rand() % 128), 128 + (PX_rand() % 128));
	for (i = 0; i < 32; i++)
	{

		pdesc->particals[i].vx = (80+randv) * PX_sin_angle(360 * i / 32.f);
		pdesc->particals[i].vy = (80+randv) * PX_cos_angle(360 * i / 32.f);
		pdesc->particals[i].clr = clr;
		pdesc->particals[i].size = 6;
	}

	clr = PX_COLOR(255, 128 + (PX_rand() % 128), 128 + (PX_rand() % 128), 128 + (PX_rand() % 128));
	for (i = 32; i < 64; i++)
	{
		pdesc->particals[i].vx = (35+randv) * PX_sin_angle(360 * (i - 32) / 32.f);
		pdesc->particals[i].vy = (35+randv) * PX_cos_angle(360 * (i - 32) / 32.f);
		pdesc->particals[i].clr = clr;
		pdesc->particals[i].size = 4;
	}

	clr = PX_COLOR(255, 128 + (PX_rand() % 128), 128 + (PX_rand() % 128), 128 + (PX_rand() % 128));
	for (i = 64; i < 96; i++)
	{
		pdesc->particals[i].vx = (15+randv) * PX_sin_angle(360 * (i - 64) / 32.f);
		pdesc->particals[i].vy = (15+randv) * PX_cos_angle(360 * (i - 64) / 32.f);
		pdesc->particals[i].clr = clr;
		pdesc->particals[i].size = 2;
	}
	pdesc->activate = PX_TRUE;
	
}

px_void PX_FireworkLauncherRenderRing(px_surface* psurface, px_float dx, px_float dy, px_float size,px_color clr)
{
	px_int x, y;
	if (size<=0)
	{
		return;
	}
	for (y = -size; y < size; y++)
		for (x = -size; x < size; x++)
		{
			if (x * x + y * y < size * size)
			{
				px_color dr=clr;
				px_float sa = 1 - sqrt(x * x + y * y) / (size);
				dr._argb.a *= sa;
				PX_SurfaceDrawPixel(psurface, dx + x, dy + y, dr);
			}

		}
}

void PX_FireworkLauncherRender(px_surface* psurface, PX_FireworkLeapLauncher* Launcher, px_dword elapsed)
{
	px_int i;
	px_float a,size,step,sx,sy;
	px_point4D scolor;
	px_dword updatetime;
	if (Launcher->elapsed+elapsed> Launcher->alive)
	{
		if (Launcher->type==PX_FIREWORK_LEAP_TYPE_NORMAL)
		{
			return;
		}
		else
		{
			if (!Launcher->Bloom.activate)
			{
				PX_FireworkBloomCreate(&Launcher->Bloom, Launcher->x, Launcher->y);		
			}
			else
			{
				PX_FireworkBloomRender(psurface, &Launcher->Bloom, elapsed);
			}
			return;
		}
	}
	
	while (elapsed)
	{
		updatetime = elapsed;
		if (elapsed > 10)
		{
			elapsed -= 10;
			updatetime = 10;
		}
		else
			elapsed = 0;

		Launcher->elapsed += updatetime;
		Launcher->reg_elapsed += updatetime;

		Launcher->x += Launcher->vx * updatetime / 1000;
		Launcher->y += Launcher->vy * updatetime / 1000;
		Launcher->vx += -Launcher->vx * updatetime / Launcher->alive;
		Launcher->vy += updatetime * 0.2f;



		a = Launcher->elapsed * 1.0f / Launcher->alive;
		size = Launcher->size*(1-a);
		
		for (i = 0; i < Launcher->reg_ipartical; i++)
		{
			Launcher->leapparticals[i].x += Launcher->leapparticals[i].vx * updatetime / 1000.f;
			Launcher->leapparticals[i].y += Launcher->leapparticals[i].vy * updatetime / 1000.f;

		}

		//Gen Partical
		while (Launcher->reg_elapsed >= PX_FIREWORKLAUNCHER_GEN_DURATION)
		{
			Launcher->reg_elapsed -= PX_FIREWORKLAUNCHER_GEN_DURATION;
			if (Launcher->reg_interpolate < PX_FIREWORKLAUNCHER_MAX_INTERPOLATE_COUNT&&size>0)
			{
				PX_memset(&Launcher->interpolate[Launcher->reg_interpolate], 0, sizeof(PX_FireworkLeapInterpolate));
				Launcher->interpolate[Launcher->reg_interpolate].x = Launcher->x;
				Launcher->interpolate[Launcher->reg_interpolate].y = Launcher->y;

				Launcher->interpolate[Launcher->reg_interpolate].color.w = Launcher->src_color.w + a * (Launcher->dst_color.w - Launcher->src_color.w);
				Launcher->interpolate[Launcher->reg_interpolate].color.x = Launcher->src_color.x + a * (Launcher->dst_color.x - Launcher->src_color.x);
				Launcher->interpolate[Launcher->reg_interpolate].color.y = Launcher->src_color.y + a * (Launcher->dst_color.y - Launcher->src_color.y);
				Launcher->interpolate[Launcher->reg_interpolate].color.z = Launcher->src_color.z + a * (Launcher->dst_color.z - Launcher->src_color.z);

				Launcher->interpolate[Launcher->reg_interpolate].size = size;

				if (Launcher->reg_ipartical<PX_COUNTOF(Launcher->leapparticals))
				{
					if (PX_rand()%15==0)
					{
						PX_memset(&Launcher->leapparticals[Launcher->reg_ipartical], 0, sizeof(PX_FireworkLeapPartical));
						Launcher->leapparticals[Launcher->reg_ipartical].x = Launcher->x;
						Launcher->leapparticals[Launcher->reg_ipartical].y = Launcher->y;

						Launcher->leapparticals[Launcher->reg_ipartical].vx = PX_randRange(-50, 50);
						Launcher->leapparticals[Launcher->reg_ipartical].vy = PX_randRange(-100,-50 );

						Launcher->leapparticals[Launcher->reg_ipartical].color = Launcher->interpolate[Launcher->reg_interpolate].color;

						Launcher->leapparticals[Launcher->reg_ipartical].size = size+PX_randRange(0,3);
						Launcher->reg_ipartical++;
					}
					
				}
				
				Launcher->reg_interpolate++;
			}
			
		}
	}


	

	step = 0;
	if (Launcher->reg_interpolate>1)
	{
		sx = Launcher->interpolate[0].x;
		sy = Launcher->interpolate[0].y;
		size = Launcher->interpolate[0].size;
		scolor = Launcher->interpolate[0].color;

		for (i = 1; i < Launcher->reg_interpolate; i++)
		{
			px_float vx = Launcher->interpolate[i].x - sx;
			px_float vy = Launcher->interpolate[i].y - sy;
			px_float distance = PX_sqrt(vx * vx + vy * vy);
			px_float change_size= Launcher->interpolate[i].size - size;
			px_point4D change_color = PX_Point4DSub(Launcher->interpolate[i].color, scolor);

			px_float dx = vx / distance;
			px_float dy = vy / distance;
			px_float dstep = Launcher->interpolate[i].size/4;
			change_color.w = Launcher->interpolate[i].color.w - scolor.w;
			scolor = Launcher->interpolate[i].color;
			step += distance;
			distance = step;
			while (step>= dstep)
			{
				PX_GeoDrawPenCircle(psurface, sx, sy, size, PX_COLOR(255*scolor.x*(1-a),255*scolor.y,255*scolor.z,255*scolor.w));
				sx += dx * dstep;
				sy += dy * dstep;
				size += dstep / distance * (change_size);
				scolor.w+= dstep / distance * (change_color.w);
				scolor.x += dstep / distance * (change_color.x);
				scolor.y += dstep / distance * (change_color.y);
				scolor.z += dstep / distance * (change_color.z);
				step -= dstep;
			}

		}

		for (i = 0; i < Launcher->reg_ipartical; i++)
		{

			PX_FireworkLauncherRenderRing(psurface, Launcher->leapparticals[i].x,
				Launcher->leapparticals[i].y,
				Launcher->leapparticals[i].size,
				PX_COLOR(\
					Launcher->leapparticals[i].color.x * 255*(1-a), \
					Launcher->leapparticals[i].color.y * 255, \
					Launcher->leapparticals[i].color.z * 255, \
					Launcher->leapparticals[i].color.w * 255  \
				)\
			);
		}

		if (Launcher->type == PX_FIREWORK_LEAP_TYPE_FLOWER)
		{
			PX_GeoDrawPenCircle(psurface, Launcher->x, Launcher->y,1.5f* Launcher->size*a, PX_COLOR(255*a, 255 * scolor.y, 255 * scolor.z, 255 * scolor.w));
		}

	}
	
}

void PX_FireworkRender(px_surface* psurface, PX_Object* pObject, unsigned int elapsed)
{
	px_int i;
	if (elapsed > 1000)elapsed = 1000;
	launcher_elapsed += elapsed;

	
	while (launcher_elapsed>= PX_FIREWORKLAUNCHER_DURATION)
	{
		px_int c;
		launcher_elapsed -= PX_FIREWORKLAUNCHER_DURATION;
		c = PX_rand() % 3+1;
		for (i = 0; i < c; i++)
		{
			PX_memset(&LeapLauncher[counter % PX_FIREWORKLAUNCHER_COUNT], 0, sizeof(PX_FireworkLeapLauncher));
			LeapLauncher[counter % PX_FIREWORKLAUNCHER_COUNT].x = psurface->width / 2 + (PX_rand() % 60) - 30.f;
			LeapLauncher[counter % PX_FIREWORKLAUNCHER_COUNT].y = psurface->height+30;
			LeapLauncher[counter % PX_FIREWORKLAUNCHER_COUNT].vx = (100 + PX_rand() % 300) * PX_cos_angle(45 + PX_rand() % 90);
			LeapLauncher[counter % PX_FIREWORKLAUNCHER_COUNT].vy = -(300.f + PX_rand() % 200) * PX_sin_angle(45 + PX_rand() % 90);

			if (PX_rand()%10==0)
			{
				LeapLauncher[counter % PX_FIREWORKLAUNCHER_COUNT].type = PX_FIREWORK_LEAP_TYPE_FLOWER;
				LeapLauncher[counter % PX_FIREWORKLAUNCHER_COUNT].src_color.x = PX_randRange(0.3, 1);
				LeapLauncher[counter % PX_FIREWORKLAUNCHER_COUNT].src_color.y = 1;
				LeapLauncher[counter % PX_FIREWORKLAUNCHER_COUNT].src_color.z = PX_randRange(0.5, 0.8);
				LeapLauncher[counter % PX_FIREWORKLAUNCHER_COUNT].src_color.w = PX_randRange(0.5, 0.8);
				LeapLauncher[counter % PX_FIREWORKLAUNCHER_COUNT].alive = PX_FIREWORKLAUNCHER_FLOWER_TIME;
			}
			else
			{
				LeapLauncher[counter % PX_FIREWORKLAUNCHER_COUNT].type = PX_FIREWORK_LEAP_TYPE_NORMAL;
				LeapLauncher[counter % PX_FIREWORKLAUNCHER_COUNT].src_color.x = PX_randRange(0.3, 1);
				LeapLauncher[counter % PX_FIREWORKLAUNCHER_COUNT].src_color.y = PX_randRange(0, 0.1);
				LeapLauncher[counter % PX_FIREWORKLAUNCHER_COUNT].src_color.z = PX_randRange(0.7, 1);
				LeapLauncher[counter % PX_FIREWORKLAUNCHER_COUNT].src_color.w = PX_randRange(0.7, 1);
				LeapLauncher[counter % PX_FIREWORKLAUNCHER_COUNT].alive = PX_FIREWORKLAUNCHER_NORMAL_TIME;
			}

			

			LeapLauncher[counter % PX_FIREWORKLAUNCHER_COUNT].dst_color.x = PX_randRange(0.5, 1);
			LeapLauncher[counter % PX_FIREWORKLAUNCHER_COUNT].dst_color.y = PX_randRange(0.5, 1);
			LeapLauncher[counter % PX_FIREWORKLAUNCHER_COUNT].dst_color.z = PX_randRange(0.5, 1);
			LeapLauncher[counter % PX_FIREWORKLAUNCHER_COUNT].dst_color.w = PX_randRange(0.5, 1);

			LeapLauncher[counter % PX_FIREWORKLAUNCHER_COUNT].size = 3+PX_rand() % 3;
			counter++;
		}
	}
	
	
	for (i=0;i<(counter>PX_FIREWORKLAUNCHER_MAX_INTERPOLATE_COUNT ? PX_FIREWORKLAUNCHER_MAX_INTERPOLATE_COUNT : counter);i++)
	{
		PX_FireworkLauncherRender(psurface, LeapLauncher + i, elapsed);
	}
	
}

static px_int desc=0;
PX_Object* PX_FireworkCreate(px_memorypool* mp, PX_Object* Parent, float x, float y)
{
	if (bRun)
	{
		return 0;
	}
	else
	{
		bRun = PX_TRUE;
	}
	return PX_ObjectCreateEx(mp, Parent, x, y, 0, 0, 0, 0, 0, 0, PX_FireworkRender, 0, &desc, sizeof(desc));
}

